/*
 * main.c - Kedrigern (Milena TTS server for Raspberry Pi)
 * Copyright (C) Bohdan R. Rau 2019 <ethanak@polip.com>
 *
 * Kedrigern is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * Kedrigern is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Kedrigern.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <signal.h>
#include <sys/wait.h>
#include <pwd.h>

#include "kedrigern_internal.h"

#include <dlfcn.h>
void *dlhandle;

struct kedrigernMouth *mouth;

struct kedrigernMouth * (*kedrigern_allocMouth)(void);
void (*kedrigern_AddParam)(struct kedrigernMouth *mouth, const char *name, const char *value);
void (*kedrigern_InitMouth)(struct kedrigernMouth *mouth);
void (*kedrigern_StartMouth)(struct kedrigernMouth *mouth, char *phonemes, double factor);
void (*kedrigern_SetParam)(struct kedrigernMouth *mouth, const char *param, const char *value);
void (*kedrigern_Free)(struct kedrigernMouth *mouth);
void (*kedrigern_ShowMouth)(struct kedrigernMouth *mouth, int millis);
void (*kedrigern_ShowPhoneme)(struct kedrigernMouth *mouth, int phoneme);
void (*kedrigern_EndSpeaking)(struct kedrigernMouth *mouth);
void (*kedrigern_AutoHide)(struct kedrigernMouth *mouth);
const char * (*kedrigern_HelpModule)(void);

char *module;

#define dlload(a) if (!(a=(void *)dlsym(dlhandle, #a))) {fprintf(stderr,"%s\n", dlerror());exit(1);}

void loadDll(char *name)
{
    char buf[256];
    sprintf(buf,MODULEDIR "/kedrigern_%s.so", name);
    if (!(dlhandle=dlopen(buf, RTLD_NOW))) {
        fprintf(stderr,"%s\n", dlerror());
        exit(1);
    }
    dlload(kedrigern_allocMouth);
    dlload(kedrigern_AddParam);
    dlload(kedrigern_InitMouth);
    dlload(kedrigern_StartMouth);
    dlload(kedrigern_SetParam);
    dlload(kedrigern_Free);
    dlload(kedrigern_ShowMouth);
    dlload(kedrigern_ShowPhoneme);
    dlload(kedrigern_EndSpeaking);
    dlload(kedrigern_AutoHide);
    dlload(kedrigern_HelpModule);
}

#define MILENA_PORT 1314

int main_socket;
struct sockaddr_in server;
struct milena *milena;
struct milena_mbrola_cfg *mbrola;
int blanked;
int autoblank = 0 ;

int guard(int n, char * err) { if (n == -1) { perror(err); exit(1); } return n; }

void initSocket(int port)
{
    main_socket=socket(AF_INET, SOCK_STREAM, 0);
    if (main_socket < 0) {
        perror("Socket");
        exit(1);
    }
    int flags = guard(fcntl(main_socket, F_GETFL), "could not get flags on TCP listening socket");
    guard(fcntl(main_socket, F_SETFL, flags | O_NONBLOCK), "could not set TCP listening socket to be non-blocking");
    int enable = 1;
    guard(setsockopt(main_socket, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(int)),
    "Setsockopt");
    server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;
    if (!port) port = MILENA_PORT;
	server.sin_port = htons( port);
	guard(bind(main_socket,(struct sockaddr *)&server , sizeof(server)), "Bind");
    guard(listen(main_socket, 5),"Listen");
}

int do_stop;
int is_speaking;

char socket_buffer[1025];

struct queue_item {
    struct queue_item *next;
    double freq;
    double tempo;
    int samplefreq;
    int realfreq;
    char str[1];
} *main_queue;

static void next_token(char **c)
{
    while (**c && !isspace(**c)) (*c)++;
    *(*c)++ = 0;
    while (**c && isspace(**c)) (*c)++;
}

static void makeQueueItem(char *cin)
{
    double tempo = 1.0;
    double freq = 1.0;
    int samplefreq = 16000;
    int realfreq = 0;

    for (;;) {
        if (!strncmp(cin,"-t",2) && isspace(cin[2])) {
            next_token(&cin);
            tempo = strtod(cin,NULL);
            next_token(&cin);
            continue;
        }
        if (!strncmp(cin,"-f",2) && isspace(cin[2])) {
            next_token(&cin);
            freq = strtod(cin,NULL);
            next_token(&cin);
            continue;
        }
        if (!strncmp(cin,"-l",2) && isspace(cin[2])) {
            next_token(&cin);
            samplefreq = strtol(cin,NULL,10);
            next_token(&cin);
            continue;
        }
        if (!strncmp(cin,"-r",2) && isspace(cin[2])) {
            next_token(&cin);
            realfreq = strtol(cin,NULL,10);
            next_token(&cin);
            continue;
        }
        break;
    }
    if (!*cin) return;
    if (realfreq == 0) {
        realfreq = samplefreq;
    }
    struct queue_item *qi = malloc(sizeof(*qi) + strlen(cin)), **qqi;
    qi->tempo = constrain(tempo,0.5,2.0);
    qi->freq = constrain(freq,0.5,2.0);
    qi->samplefreq = constrain(samplefreq, 8000, 32000);
    qi->samplefreq -= qi->samplefreq % 50;
    qi->realfreq = constrain(realfreq, 8000, 32000);
    qi->realfreq -= qi->realfreq % 50;
    strcpy(qi->str, cin);
    for (qqi=&main_queue; *qqi; qqi = &(*qqi)->next);
    *qqi = qi;
    qi->next = NULL;
}

void socketCommand(int client_socket)
{
    
    void setMouth(char *param,int n)
    {
        char buf[32],*c;
        if (!module || !kedrigern_SetParam) return;
        sprintf(buf,"%s:%s",module, param);
        sprintf(buf+strlen(buf)+1, "%d", n);
        kedrigern_SetParam(mouth, buf, buf+strlen(buf)+1);
    }
    int len = recv(client_socket, socket_buffer, 1024, 0);
    if (len <= 0) return;
    socket_buffer[len] = 0;
    char *cin = socket_buffer;
    char *cmd = cin;
    next_token(&cin);
    if (!strcmp(cmd,"stop")) {
        do_stop = 1;
        send(client_socket, "OK", 2, 0);
        return;
    }
    if (!strcmp(cmd,"bright")) {
        if (*cin) setMouth("brightness", atoi(cin));
        send(client_socket, "OK", 2, 0);
        return;
    }
    if (!strcmp(cmd,"phonem")) {
        if (*cin && kedrigern_ShowPhoneme) kedrigern_ShowPhoneme(mouth, atoi(cin));
        //send(client_socket, "OK", 2, 0);
        return;
    }
    if (!strcmp(cmd,"status")) {
        if (is_speaking) send(client_socket, "busy", 4, 0);
        else send(client_socket, "idle", 4, 0);
        return;
    }
    
    if (!strcmp(cmd,"speak")) {
        makeQueueItem(cin);
        send(client_socket, "OK", 2, 0);
        return;
    }
    send(client_socket, "ERR", 3, 0);
    //printf("Command is <%s>\n", cmd);
    //printf("Params are <%s>\n", cin);
    
}

int tombrola[2];
int frommbrola[2];
pid_t mbrola_pid;
int play_time;
int bufsize;

void doSpeaking(void)
{
    char bufor[bufsize];
    int n=read(frommbrola[0],bufor, bufsize);
    if (n <= 0) {
        finalizeAudio();
        if (kedrigern_EndSpeaking) kedrigern_EndSpeaking(mouth);
        close(frommbrola[0]);
        is_speaking = 0;
        
        return;
    }
    
    if (kedrigern_ShowMouth) kedrigern_ShowMouth(mouth, play_time);
    play_time += 10;
    playSound(bufor, n);
}

void initSpeak(struct queue_item *qi)
{
    char *pho = convertMilena(qi->str);
    if (!pho) return;
    //printf("pho %s\n", pho);
    pipe(tombrola);
    pipe(frommbrola);
    mbrola_pid=fork();
    if (mbrola_pid < 0) {
        perror("fork");
        exit(1);
    }
    if (!mbrola_pid) {
        close(tombrola[1]);
        dup2(tombrola[0], 0);
        close(frommbrola[0]);
        dup2(frommbrola[1], 1);
        char cmpt[8], cmpf[8], cmps[8];
        sprintf(cmpt,"%.2f", (qi->tempo * qi->realfreq) / qi->samplefreq);
        sprintf(cmpf,"%.2f", qi->freq);
        sprintf(cmps,"%d", qi->samplefreq);
        execl("/usr/bin/mbrola","mbrola","-e", 
        "-t",cmpt, "-f", cmpf, "-l", cmps,"-v", "1.4",
        "/usr/share/mbrola/pl1/pl1","-","-",NULL);
        exit(0);
    }
    close(frommbrola[1]);
    close(tombrola[0]);
    FILE *fmb = fdopen(tombrola[1],"w");
    if (!fmb) {
        perror("Dup2");
        exit(1);
    }
    if (kedrigern_StartMouth) {
        kedrigern_StartMouth(mouth, pho, qi->tempo); // (qi->tempo * qi->realfreq) / qi->samplefreq);
    }
    fputs(pho,fmb);
    fclose(fmb);
    free(pho);
    startAudio(qi->realfreq);
    bufsize = qi->realfreq / 50;
    is_speaking = 1;
    play_time = 0;
    
}



void help(char *s, int xit)
{
    fprintf(xit?stderr:stdout,"Kedrigern V" VERSION "\n\
Sposób użycia:\n\
    %s -L\n\
    %s [-m moduł] [-C <plik default>] [ocje] [opcje modułu]\n\
gdzie opcje to:\n\
-U <słownik wymowy dla Mileny>\n\
-p <port>\n\
-D <nazwa urządzenia audio>\n\
Podanie -L powoduje wypisanie możliwych urządzeń audio i wyjście\n\
W przypadku kilku parametrów -m liczy się pierwszy\n\
W przypadku kilku parametrów -m liczy się pierwszy\n\
Podanie pustego stringu jako moduł wyłącza animację ust\n", s, s);
    if (kedrigern_HelpModule) fprintf(stderr,"%s",kedrigern_HelpModule());
    exit(xit);
}


static char * strdups(const char *d)
{
    int len,i;
    while (*d && isspace(*d)) d++;
    for (len = i = 0; d[i]; i++) if (!isspace(d[i])) len = i+1;
    if (len == 0) return NULL;
    char *rc = malloc(len+1);
    memcpy(rc,d,len);
    rc[len]=0;
    return rc;
}

static char *trim(char *c)
{
    char *d, *e;
    while (*c && isspace(*c)) c++;
    for (e=d=c; *d; d++) if (!isspace(*d)) e=d+1;
    *e=0;
    return c;
}

    
int main(int argc, char *argv[])
{
    int port=0, opt;
    char *audiodev=NULL, *helo=NULL, *shape=NULL, *udic=NULL;
        
    void defaultFile(char *fname)
    {
        FILE *f=fopen(fname,"r");
        if (!f) return;
        char buf[256];
        while(fgets(buf,256,f)) {
            char *c=buf,*d;
            while(*c && isspace(*c)) c++;
            if (!*c || *c=='#') continue;
            d=strchr(c,'=');
            if (!d) continue;
            *d++=0;
            if (!strcasecmp(c,"PORT")) port = strtol(d, NULL, 10);
            else if (!strcasecmp(c,"MODULE")) {
                d=trim(d);
                if (!module) {
                    module = d;
                    if (*module) {
                        loadDll(module);
                        mouth = kedrigern_allocMouth();
                    }
                }
            }
            else if (!strcmp(c,"HELO")) helo=strdups(d);
            else if (!strcmp(c,"DICT")) udic=strdups(d);
            else {
                if (kedrigern_AddParam) kedrigern_AddParam(mouth, c, d);
            }
        }
        fclose(f);
    }
            
        
    while ((opt = getopt(argc, argv, "p:hC:LD:U:m:M:")) != -1) {
        switch(opt) {
            case 'M':
            if (!module) {
                module = optarg;
                if (*module) {
                    loadDll(module);
                    mouth = kedrigern_allocMouth();
                }
            }
            break;
            
            case 'm':
            if (mouth && kedrigern_AddParam) kedrigern_AddParam(mouth, optarg, NULL);
            break;
            
            case 'C':
            defaultFile(optarg);
            break;
            
            case 'p':
            port = atoi(optarg);
            break;
                        
            case 'h':
            help(argv[0],0);
            break;
            
            
            case 'D':
            audiodev = optarg;
            break;
            
            case 'L':
            showAudioDrivers();
            exit(0);
            
            case 'U':
            udic = optarg;
            break;

            default:
            help(argv[0],1);
        }
    }
    
    initAudio(audiodev);
    initSocket(port);
    initMilena(udic);
    if (kedrigern_InitMouth) kedrigern_InitMouth(mouth);
    if (helo) makeQueueItem(helo);
    
    for (;;) {
        int waitme = 0;
        int client_socket = accept(main_socket, NULL, NULL);
        if (client_socket < 0) {
            if (errno == EWOULDBLOCK) waitme = 1;
            else {
                perror("Accept");
                exit(1);
            }
        }
        else {
            socketCommand(client_socket);
            close(client_socket);
        }
        if (do_stop) {
            do_stop = 0;
            struct queue_item *qi;
            while (qi=main_queue) {
                main_queue = qi->next;
                free(qi);
            }
            if (is_speaking) {
                finalizeAudio();
                if (kedrigern_EndSpeaking) kedrigern_EndSpeaking(mouth);
                close(frommbrola[0]);
                kill(mbrola_pid, SIGKILL);
                waitpid(mbrola_pid,NULL, 0);
                is_speaking = 0;
            }
            continue;
        }
        if (!is_speaking && !main_queue) {
            if (waitme) usleep(10000);
            continue;
        }
        if (is_speaking) {
            doSpeaking();
            continue;
        }
        if (!main_queue) continue;
        struct queue_item *qi = main_queue;
        main_queue = qi->next;
        initSpeak(qi);
        free(qi);
            
    }
    return 0;
}
